/*
 * $Id: GeoPosCorrectorVPI.java,v 1.11 2005/03/25 00:13:58 hartmut Exp $
 *
 * Copyright (C) 2002 by Brockmann Consult (info@brockmann-consult.de)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation. This program is distributed in the hope it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.bc.beam.geocorr;

import com.bc.progress.ProgressController;
import com.bc.progress.ProgressControllerPool;
import com.bc.swing.ProgressMonitor;
import org.esa.beam.framework.datamodel.*;
import org.esa.beam.framework.dataop.dem.*;
import org.esa.beam.framework.ui.SwingWorker;
import org.esa.beam.framework.ui.command.CommandEvent;
import org.esa.beam.framework.ui.command.CommandListener;
import org.esa.beam.framework.ui.command.CommandManager;
import org.esa.beam.framework.ui.command.ExecCommand;
import org.esa.beam.visat.VisatApp;
import org.esa.beam.visat.VisatPlugIn;

public class GeoPosCorrectorVPI implements VisatPlugIn {

    public void initPlugIn(final VisatApp visatApp) {
        final CommandManager commandManager = visatApp.getCommandManager();
        final CommandListener commandListener = new CommandListener() {
            public void actionPerformed(CommandEvent event) {
                computeCorrections(visatApp);
            }

            public void updateState(CommandEvent event) {
                final Product selectedProduct = visatApp.getSelectedProduct();
                final boolean enabled = selectedProduct != null &&
                        selectedProduct.getGeoCoding() != null &&
                        selectedProduct.getGeoCoding().canGetPixelPos() &&
                        selectedProduct.canBeOrthorectified();
                event.getCommand().setEnabled(enabled);
            }
        };
        final ExecCommand command = commandManager.createExecCommand("GeoPosCorrectorVPI", commandListener);
        command.setParent("Tools");
        command.setText("Compute Geodetic Corrections");
        command.setLongDescription("Computes the orthorectified geodetic corrections");
        command.setShortDescription("Computes the orthorectified geodetic corrections and adds it to the selected product");

//        visatApp.addHelp(this, "com/bc/beam/geocorr/help/GeoPosCorrector.hs");
    }

    public void updateComponentTreeUI() {

    }

    private void computeCorrections(final VisatApp visatApp) {
        final String demName = "GETASSE30";
        final ElevationModelDescriptor demDescriptor = ElevationModelRegistry.getInstance().getDescriptor(demName);
        if (demDescriptor == null || !demDescriptor.isDemInstalled()) {
            visatApp.showErrorDialog("DEM not ready: " + demName);
            return;
        }

        final Product source = visatApp.getSelectedProduct();
        final Product subset = source;

        if (source.getNumBands() == 0 || !source.getBandAt(0).canBeOrthorectified()) {
            visatApp.showErrorDialog("Orthorectification is not possible for this product.");
            return;
        }

        final boolean useOrthorectifier2 = false;
        final ProgressMonitor pl = new ProgressMonitor(visatApp.getMainFrame(), "Computing Geodetic Corrections");

        final SwingWorker swingWorker = new SwingWorker(pl) {
            public Object construct() {

                final GeoCoding geoCoding = subset.getGeoCoding();
                final int width = subset.getSceneRasterWidth();
                final int height = subset.getSceneRasterHeight();

                final Band demElevationBand = addBand(subset, width, height, "dem_elevation", "m", "DEM " + demName + " elevation");
                final Band orthoCorrLatBand = addBand(subset, width, height, "ortho_corr_lat", "degree", "Orthorectification corrected latitude");
                final Band orthoCorrLonBand = addBand(subset, width, height, "ortho_corr_lon", "degree", "Orthorectification corrected longitude");
                final Band orthoDeltaXBand = addBand(subset, width, height, "ortho_delta_x", "pixel", "Orthorectifified (false) pixel delta in X direction");
                final Band orthoDeltaYBand = addBand(subset, width, height, "ortho_delta_y", "pixel", "Orthorectifified (false) pixel delta in Y direction");
                addVirtualBand(subset, width, height, "ortho_delta_lat", "degree", "ortho_corr_lat - latitude");
                addVirtualBand(subset, width, height, "ortho_delta_lon", "degree", "ortho_corr_lon - longitude");

                final ElevationModel dem = demDescriptor.createDem();
                final Pointing pointing = source.getBandAt(0).getPointing();
                final Orthorectifier orthorectifier = new Orthorectifier(width, height, pointing, dem, 100);

                Band ortho2DeltaXBand = null;
                Band ortho2DeltaYBand = null;
                Orthorectifier orthorectifier2 = null;
                if (useOrthorectifier2) {
                    ortho2DeltaXBand = addBand(subset, width, height, "ortho2_delta_x", "pixel", "Orthorectifified-2 (false) pixel delta in X direction");
                    ortho2DeltaYBand = addBand(subset, width, height, "ortho2_delta_y", "pixel", "Orthorectifified-2 (false) pixel delta in Y direction");
                    addVirtualBand(subset, width, height, "dx_minus_dx2", "pixel", "ortho_delta_x - ortho2_delta_x");
                    addVirtualBand(subset, width, height, "dy_minus_dy2", "pixel", "ortho_delta_y - ortho2_delta_y");
                    orthorectifier2 = new Orthorectifier2(width, height, pointing, dem, 100);
                }

                final PixelPos sourcePixelPos = new PixelPos();
                final GeoPos sourceGeoPos = new GeoPos();
                final PixelPos orthoCorrPixelPos = new PixelPos();
                final PixelPos ortho2CorrPixelPos = new PixelPos();
                final GeoPos orthoCorrGeoPos = new GeoPos();
                float demElevation;

                final ProgressController pc = ProgressControllerPool.getInstance().getProgressController();
                pc.fireProcessStarted("Computing geodetic corrections...", 0, height - 1);
                try {
                    for (int y = 0; y < height; y++) {
                        for (int x = 0; x < width; x++) {
                            sourcePixelPos.setLocation(x + 0.5f, y + 0.5f);
                            geoCoding.getGeoPos(sourcePixelPos, sourceGeoPos);
                            orthorectifier.getGeoPos(sourcePixelPos, orthoCorrGeoPos);
                            orthorectifier.getPixelPos(sourceGeoPos, orthoCorrPixelPos);
                            try {
                                demElevation = dem.getElevation(orthoCorrGeoPos);
                            } catch (Exception e) {
                                demElevation = dem.getDescriptor().getNoDataValue();
                            }
                            demElevationBand.setPixelFloat(x, y, demElevation);
                            orthoCorrLatBand.setPixelFloat(x, y, orthoCorrGeoPos.lat);
                            orthoCorrLonBand.setPixelFloat(x, y, orthoCorrGeoPos.lon);
                            orthoDeltaXBand.setPixelFloat(x, y, orthoCorrPixelPos.x - sourcePixelPos.x);
                            orthoDeltaYBand.setPixelFloat(x, y, orthoCorrPixelPos.y - sourcePixelPos.y);
                            if (useOrthorectifier2) {
                                orthorectifier2.getPixelPos(sourceGeoPos, ortho2CorrPixelPos);
                                ortho2DeltaXBand.setPixelFloat(x, y, ortho2CorrPixelPos.x - sourcePixelPos.x);
                                ortho2DeltaYBand.setPixelFloat(x, y, ortho2CorrPixelPos.y - sourcePixelPos.y);
                            }
                        }
                        pc.fireProcessInProgress(y);
                    }
                } finally {
                    pc.fireProcessEnded();
                }
                return null;
            }

            public void finished() {
                visatApp.getProductManager().removeProduct(subset);
                visatApp.getProductManager().addProduct(subset);
            }
        };
        swingWorker.start();
    }

    private static void addVirtualBand(final Product subset, final int width, final int height, final String name, final String unit, final String expression) {
        final VirtualBand band = new VirtualBand(name, ProductData.TYPE_FLOAT32, width, height, expression);
        band.setUnit(unit);
        band.setDescription(expression);
        subset.addBand(band);
    }

    private static Band addBand(final Product subset, final int width, final int height, final String bandName, final String unit, final String description) {
        final Band band = new Band(bandName, ProductData.TYPE_FLOAT32, width, height);
        band.setUnit(unit);
        band.setDescription(description);
        band.setSynthetic(true);
        subset.addBand(band);
        band.ensureRasterData();
        return band;
    }
}
